import scene
from defs import *
from numeric import *

class Camera(scene.Item):
    def __init__(self, center, at, up, fov, size, near=1.0, far=100000.0):
        self.lookat(center, at, up)
        self.fov = float(fov)
        self.size = size
        w, h = map(float, size)
        self.perspective(fov*DTOR, w/h, near, far)
        self.viewport((0, 0), (w, h))

    def lookat(self, c, a, u):
        self.at = array(a, float)
        self.center = array(c, float)
        to = self.at - self.center
        self.set_normal(to)
        self.set_up(array(u, float))
        self.makeV()

    def perspective(self, fov, ar, near, far):
        if ar < EPSILON: raise 'illegal aspect ratio'
        self.set_distance(near)
        self.set_depth(near, far)
        center = (0, 0)
        size = tan(fov/2.0)*near, tan(fov/2.0)*near/ar
        self.set_window(center, size)
        self.makeC()
        self.makeP()

    def viewport(self, center, size):
        x, y = center
        w, h = size
        self.set_viewport(x, y, -32768.0, x+w, y+h, 32767.0)
        self.makeS()

    def resize(self, size):
        w, h = map(float, size)
        self.size = size
        self.perspective(self.fov*DTOR, w/h, self.near, self.far)
        self.viewport((0, 0), (w, h))

    def set_normal(self, to):
        if norm(to) < EPSILON: raise 'invalid view plane normal'
        self.normal = to/norm(to)

    def set_up(self, up):
        if norm(up) < EPSILON: raise 'no view up direction'
        self.up = up

    def set_distance(self, dist):
        if abs(dist) < EPSILON: raise 'invalid view distance'
        self.dist = dist

    def set_depth(self, near, far):
        if abs(far - near) + abs(far) < 2*EPSILON: raise 'invalid view depth'
        self.near = near
        self.far = far

    def set_window(self, center, size):
        if abs(size[0]) + abs(size[1]) < 2*EPSILON: raise 'invalid window size'
        self.wcenter = center
        self.wsize = size

    def set_viewport(self, a, b, c, x, y, z):
        if abs(a-x) + abs(b-y) < EPSILON: raise 'invalid viewport'
        self.smin = (a, b, c)
        self.smax = (x, y, z)

    def makeV(self):
        n = self.normal
        v = self.up - (dot(self.up,n)*n)
        if norm(v) < EPSILON: raise 'view up parallel to view normal'
        v = v / norm(v)
        u = cross(n, v)
        V = identity(4)
        c = self.center
        t = array([dot(c, u), dot(c, v), dot(c, n)])
        V[0][0:3] = u
        V[1][0:3] = v
        V[2][0:3] = n
        V[:,3][0:3] = -t
        self.V = V

    def makeC(self):
        C = identity(4)
        b, d = self.far, self.dist
        su, sv = self.wsize
        cu, cv = self.wcenter
        C[0,0] = d / (su*b)
        C[1,1] = d / (sv*b)
        C[2,2] = 1.0 / b
        C[0,2] = -cu / (su*b)
        C[1,2] = -cv / (sv*b)
        self.C = C

    def makeP(self):
        P = identity(4)
        near, far = self.near, self.far
        f = 1.0 / (far - near)
        P[2,2] = far*f
        P[2,3] = -near*f
        P[3,2] = 1
        P[3,3] = 0
        self.P = P

    def makeS(self):
        S = identity(4)
        a, b, c = self.smin
        x, y, z = self.smax
        S[0,0] = (x-a)/2.0
        S[1,1] = (y-b)/-2.0 # corrects y in monitor
        S[2,2] = z-c
        S[0,3] = (x+a+1.0)/2.0
        S[1,3] = (y+b+1.0)/2.0
        S[2,3] = c + 0.5
        self.S = S

    def transformations(self):
        return self.V, self.C, self.P, self.S

    def rotate(self, theta=0.0, phi=0.0, psi=0.0):
        R1 = rotation(theta, 0)
        R2 = rotation(phi, 1)
        R = dot(R1, R2)
        c, a, u, V = self.center, self.at, self.up, self.V
        t = c + u
        a2 = dot(inv(V), dot(R, dot(V, hcoord(a))))[0:3]
        t2 = dot(inv(V), dot(R, dot(V, hcoord(t))))[0:3]
        c += a - a2
        self.lookat(c, a, t2-c)

def hcoord(v):
    h = ones(4)
    h[0:3] = v
    return h

def rotation(angle, axis):
    if axis == 0: i, j = 1, 2
    elif axis == 1:  i, j = 0, 2
    elif axis == 2:  i, j = 0, 1
    else: raise 'Invalid axis'
    R = identity(4)
    R[i,i] = cos(angle)
    R[i,j] = sin(angle)
    R[j,i] = -sin(angle)
    R[j,j] = cos(angle)
    return R

