from defs import *
from numeric import *
from time import time
from objects.mesh import Mesh, Vertex, Edge, Face
from ray import Range


TREE_GRANULARITY = 2**4
TREE_H_MAX = 30
EPS = 0 #1e-15
h = 0


class Node(object):
    def __init__(self, parent, center, dimension):
        self.parent = parent
        self.c = center
        self.dim = dimension # canonic dimension
        self.H1 = None # (+) positive hyperspace
        self.H2 = None # (-) negative hyperspace

def construct_kd_tree(mesh):
    bounds = mesh.box.ll.copy(), mesh.box.ur.copy()
    root, h = create_node(mesh.faces, [], bounds, None)
    print 'h:', h
    return root

def create_node(faces, subfaces, bounds, parent):
    semifaces = list(set([sf.parent for sf in subfaces]))
    n = len(faces) + len(semifaces)
    if n <= TREE_GRANULARITY:
        return faces + semifaces, 0

    m, dim, (F1, SF1, b1), (F2, SF2, b2) = split(faces, subfaces, bounds)
    node = Node(parent, m, dim)
    node.H1, h1 = create_node(F1, SF1, b1, node)
    node.H2, h2 = create_node(F2, SF2, b2, node)
    return node, max(h1, h2) + 1

def is_subface(face):
    return face.id < 0

def split(faces, subfaces, bounds):
    m, dim = split_location(faces, subfaces, bounds)
    b1, b2 = split_bounds(m, dim, bounds)
    F1, F2, SF1, SF2 = [], [], [], []
    for f in faces + subfaces:
        bin1, bin2 = test_face(f, m, dim)
        n1 = len(bin1)
        if n1 == 3:
            if is_subface(f): SF1.append(f)
            else: F1.append(f)
        elif n1 == 2:
            A, (B, C) = split_triangle(m, dim, f, bin2[0], bin1, -EPS)
            SF2.append(A)
            SF1.append(B)
            SF1.append(C)
        elif n1 == 1:
            A, (B, C) = split_triangle(m, dim, f, bin1[0], bin2, EPS)
            SF1.append(A)
            SF2.append(B)
            SF2.append(C)
        else:
            if is_subface(f): SF2.append(f)
            else: F2.append(f)
    return m, dim, (F1, SF1, b1), (F2, SF2, b2)

def test_face(face, m, dim):
    bin1, bin2 = [], []
    for v in face.vertices:
        if (v.vector[dim]-m) >= 0: bin1.append(v)
        else: bin2.append(v)
    return bin1, bin2

def split_triangle(m, dim, face, cv, (vi, vj), offset):
    Id = identity(3)
    ti = (m - vi.vector[dim]) / (cv.vector[dim] - vi.vector[dim])
    tj = (m - vj.vector[dim]) / (cv.vector[dim] - vj.vector[dim])
    wi = cv.vector + ti*(vi.vector - cv.vector)
    wj = cv.vector + tj*(vj.vector - cv.vector)
    ###
    o = Id[dim]*offset
    A = Face([cv, Vertex(wi+o, -1), Vertex(wj+o, -1)], -1)
    Wi = Vertex(wi-o, -1)
    Wj = Vertex(wj-o, -1)
    B = Face([Wi, vi, vj], -1)
    C = Face([Wj, Wi, vj], -1)
    if is_subface(face): parent = face.parent
    else: parent = face
    A.parent = parent; A.center = mean(A.vectors, 0)
    B.parent = parent; B.center = mean(B.vectors, 0)
    C.parent = parent; C.center = mean(C.vectors, 0)
    return A, (B, C)

def split_location(faces, subfaces, bounds):
    centers = [f.center for f in faces + subfaces]
    C = array(centers)
    l = zip(var(C, 0), range(3))
    l.sort()
    dim = l[-1][1]
    m = mean(C, 0)[dim]
    return m, dim

def split_bounds(m, dim, bounds):
    ll, ur = map(copy, bounds)
    ll[dim] = m
    ur[dim] = m
    b2 = bounds[0], ur
    b1 = ll, bounds[1]
    return b1, b2

class SearchContext(object):
    def __init__(self):
        self.stack = []
        self.hits = []

    def hit(self):
        if len(self.hits) > 0:
           self.hits.sort()
           return self.hits[0][1]

def kd_search(ray, mesh):
    context = SearchContext()
    intersect = lambda x, y: x & y
    I = reduce(intersect, ray.box_intersection_ranges(mesh.box))
    search_kd_node(context, mesh.kd_tree, ray, I)
    return context.hit()

def search_kd_node(context, node, ray, I):
    if type(node) is not Node:
        search_kd_leaf(context, node, ray, I)
    else:
        search_kd_split(context, node, ray, I)

def search_kd_split(context, node, ray, I):
    dim = node.dim
    q = node.c
    p, v = ray.p[dim], ray.v[dim]
    first, second = kd_sub_nodes(node, v)
    t = (q-p)/v
    if t >= I.b:
        search_kd_node(context, first, ray, I)
    elif t <= I.a:
        search_kd_node(context, second, ray, I)
    else:
        context.stack.append((second, Range(t, I.b)))
        search_kd_node(context, first, ray, Range(I.a, t))

def kd_sub_nodes(node, v):
    if v >= 0:
        return node.H2, node.H1
    else:
        return node.H1, node.H2

def search_kd_leaf(context, leaf, ray, I):
    t, hit = search_kd_hit(leaf, ray)
    if (hit is not None) and t <= I.b:
        context.hits += [ (t, hit) ]
    else:
        continue_kd_search(context, ray)

def search_kd_hit(faces, ray):
    hits = []
    for triangle in faces:
        t = ray.intersect_triangle(triangle.vectors, triangle.normal)
        if t is not None and t > 0: return t, triangle
    return -1, None

def continue_kd_search(context, ray):
    if len(context.stack) > 0:
        node, I = context.stack.pop()
        search_kd_node(context, node, ray, I)


