import sdl
from numeric import *
from time import time
from defs import *
from ray import *

def project(T, v):
    u = ones(4)
    u[0:3] = v
    p = dot(T, u)
    f = 1/p[3]
    p *= f
    return p[0], p[1], p[2]

def roundval(values):
    return map(int, values)

def is_backface(v, n, cam):
    e = v - cam.center
    return dot(n, e) < -0.1

def make_transform(cam):
    V, C, P, S = cam.transformations()
    T = dot(S, dot(P, dot(C, V)))
    return T

def make_inverse_transform(cam):
    Vi, Ci, Pi, Si = map(inv, cam.transformations())
    Ti = dot(Vi, dot(Ci, dot(Pi, Si)))
    return Ti

def make_ray(cam, (u, v)):
    Ti = make_inverse_transform(cam)
    w = dot(Ti, array([u, v, cam.smax[2], 1.0]))
    c = dot(Ti, array([0.0, 0.0, 1.0, 0.0]))
    v = w[0:3]/w[3]
    v /= norm(v)
    return Ray(cam.center, v)

def point_render(obj, cam, painter, color):
    T = make_transform(cam)
    for v in obj.vectors:
        x, y = project(T, v)[0:2]
        painter.paint(x, y, color)

def wireframe_render(obj, cam, surface, color):
    T = make_transform(cam)
    proj = lambda v: project(T, v)[0:2]
    for face in obj.faces:
        poly = map(proj, face.vectors)
        sdl.draw.aalines(surface, color, True, poly)

def hidden_wireframe_render(obj, cam, surface, color):
    T = make_transform(cam)
    proj = lambda v: project(T, v)[0:2]
    for face in obj.faces:
        if is_backface(face.center, face.normal, cam): continue
        poly = map(proj, face.vectors)
        sdl.draw.aalines(surface, color, True, poly)


def paint_poly(poly, cam, surface, color=[0, 0, 0]):
    T = make_transform(cam)
    proj = lambda v: project(T, v)[0:2]
    for i in range(len(poly)):
        x0, y0 = proj(poly[i-1])
        x1, y1 = proj(poly[i])
        sdl.draw.line(surface, color, (x0, y0), (x1, y1))
        sdl.display.flip()

####################

GRANULARITY=250

def render_depth_map(obj, cam, painter, wait):
    raster_mesh(obj, cam, painter, raster_triangle_depth, wait)

def render_normal_map(obj, cam, painter, wait):
    raster_mesh(obj, cam, painter, raster_triangle_normal, wait)

def raster_mesh(obj, cam, painter, action, wait):
    c = 0
    T = make_transform(cam)
    t = float(len(obj.faces))
    for vertex in obj.vertices:
        vertex.projection = project(T, vertex.vector)
    for triangle in obj.faces:
        c += 1
        if c%GRANULARITY == 1:
            wait(c/t)
        if is_backface(triangle.center, triangle.normal, cam):
            continue
        sort_triangle_vertices(obj, triangle)
        action(obj, triangle, painter)

def sort_triangle_vertices(obj, triangle):
    a, b, c = triangle.vertices
    ay, by, cy = [v.projection[1] for v in triangle.vertices]
    if ay < by:
        if by < cy: order = a, b, c
        elif cy < ay: order = c, a, b
        else: order = a, c, b
    else:
        if cy < by: order = c, b, a
        elif ay < cy: order = b, a, c
        else: order = b, c, a
    triangle.vertices = order

def raster_triangle_depth(obj, triangle, painter):
    v1, v2, v3 = [v.projection for v in triangle.vertices]
    y1, y2, y3 = [int(v[1]) for v in v1, v2, v3]
    h = (y3 - y1) + 1
    ha = (y2 - y1) + 1
    hb = (y3 - y2) + 1
    X, Y, Z = rasterXYZ(v1, v3, h)
    Xa, Ya, Za = rasterXYZ(v1, v2, ha)
    Xb, Yb, Zb = rasterXYZ(v2, v3, hb)

    for y, x0, x1, z0, z1 in zip(Ya, X[0:ha], Xa, Z[0:ha], Za):
        w = abs(x1-x0)+1
        for x, z in zip(linspace(x0, x1, w), linspace(z0, z1, w)):
            painter.dpaint(x, y, z)
    for y, x0, x1, z0, z1 in zip(Yb, X[ha-1:], Xb, Z[ha-1:], Zb):
        w = abs(x1-x0)+1
        for x, z in zip(linspace(x0, x1, w), linspace(z0, z1, w)):
            painter.dpaint(x, y, z)

def raster_triangle_normal(obj, triangle, painter):
    v1, v2, v3 = [v.projection for v in triangle.vertices]
    n1, n2, n3 = [v.normal for v in triangle.vertices]
    y1, y2, y3 = [int(v[1]) for v in v1, v2, v3]
    h =  (y3 - y1) + 1
    ha = (y2 - y1) + 1
    hb = (y3 - y2) + 1
    X, Y, Z       = rasterXYZ(v1, v3, h)
    Xa, Ya, Za    = rasterXYZ(v1, v2, ha)
    Xb, Yb, Zb    = rasterXYZ(v2, v3, hb)
    NX, NY, NZ    = raster(n1, n3, h)
    NXa, NYa, NZa = raster(n1, n2, ha)
    NXb, NYb, NZb = raster(n2, n3, hb)

    for i in range(ha):
        y = Ya[i]
        x0, x1 = X[i], Xa[i]
        z0, z1 = Z[i], Za[i]
        nx0, nx1 = NX[i], NXa[i]
        ny0, ny1 = NY[i], NYa[i]
        nz0, nz1 = NZ[i], NZa[i]
        w = abs(x1-x0) + 1
        XX = linspace(x0, x1, w).astype(int)
        ZZ = linspace(z0, z1, w)
        NN = zip(linspace(nx0, nx1, w),
                 linspace(ny0, ny1, w),
                 linspace(nz0, nz1, w))
        for x, z, n in zip(XX, ZZ, NN):
            painter.npaint(x, y, z, n)

    for i in range(hb):
        y = Yb[i]
        p = ha-1
        x0, x1 = X[p+i], Xb[i]
        z0, z1 = Z[p+i], Zb[i]
        nx0, nx1 = NX[p+i], NXb[i]
        ny0, ny1 = NY[p+i], NYb[i]
        nz0, nz1 = NZ[p+i], NZb[i]
        w = abs(x1-x0) + 1
        XX = linspace(x0, x1, w).astype(int)
        ZZ = linspace(z0, z1, w)
        NN = zip(linspace(nx0, nx1, w),
                 linspace(ny0, ny1, w),
                 linspace(nz0, nz1, w))
        for x, z, n in zip(XX, ZZ, NN):
            painter.npaint(x, y, z, n)

def rasterXYZ(A, B, n):
    X = linspace(A[0], B[0], n).astype(int)
    Y = linspace(A[1], B[1], n).astype(int)
    Z = linspace(A[2], B[2], n)
    return X, Y, Z

def raster(A, B, n):
    X = linspace(A[0], B[0], n)
    Y = linspace(A[1], B[1], n)
    Z = linspace(A[2], B[2], n)
    return X, Y, Z
