#!/usr/bin/python
import sdl
from scene import read
from sys import argv
from time import time
from viewer import Viewer
from numeric import *
from utility import *
from render import *
from ray import *
from kd_tree import *
from image_space import *
from object_space import *


TAU = 0.5
RATE = 0.005
ROT = 15

class SceneViewer(Viewer):
    def __init__(self, scene, caption):
        Viewer.__init__(self, scene.cameras[0].size, caption)
        self.scene = scene
        self.camera = self.scene.cameras[0]
        self.obj = self.scene.objects[0]
        self.fgcolor = (0, 0, 0)
        self.bgcolor = (255, 255, 255)
        self.clear()
        self.precalcs()
        self.fast_display()
        self.drx, self.dry, self.dt = 0.0, 0.0, 0
        self.eye = self.camera.center
        self.shots = 0

    def precalcs(self):
        self.message('precalc: bounding box...')
        self.bbox = compute_bbox(self.obj)
        self.mean = compute_mean(self.obj)
        self.message('precalc: face normals...')
        compute_face_normals(self.obj)
        self.message('precalc: adjacent faces...')
        find_adjacent_faces(self.obj)
        self.message('precalc: vertex normals...')
        compute_vertex_normals(self.obj)
        self.message('precalc: edges...')
        create_edges(self.obj)
        self.obj.box = Box.compute(self.obj)
        self.message('precalc: building kd-tree...')
        self.construct_kd_tree()

    def on_key_down(self, event):
        if event.key == sdl.K_ESCAPE:
            self.stop()

        elif event.key in [sdl.K_LEFT, sdl.K_KP4]:
            self.camera.rotate(0.0, DTOR*ROT)
            self.fast_display()

        elif event.key in [sdl.K_RIGHT, sdl.K_KP6]:
            self.camera.rotate(0.0, DTOR*-ROT)
            self.fast_display()

        elif event.key in [sdl.K_UP, sdl.K_KP8]:
            self.camera.rotate(DTOR*ROT)
            self.fast_display()

        elif event.key in [sdl.K_DOWN, sdl.K_KP2]:
            self.camera.rotate(DTOR*-ROT)
            self.fast_display()

        elif event.key == sdl.K_w:
            self.wireframe()

        elif event.key == sdl.K_p:
            self.points()

        elif event.key == sdl.K_d:
            self.depth_map()

        elif event.key == sdl.K_n:
            self.normal_map()

        elif event.key == sdl.K_e:
            self.eye = self.camera.center
            'eye:', self.eye

        elif event.key == sdl.K_1:
            self.compute_silhouette()

        elif event.key == sdl.K_F1:
            self.construct_kd_tree()

        elif event.key == sdl.K_F2:
            self.show_kd_tree()

        elif event.key == sdl.K_F5:
            self.detect_dmap_edges()

        elif event.key == sdl.K_F6:
            self.detect_nmap_edges()

        elif event.key == sdl.K_F7:
            self.combine_dmap_nmap()

        elif event.key == sdl.K_F11:
            sdl.display.toggle_fullscreen()

        elif event.key == sdl.K_F12:
            self.screenshot()

        elif event.key == sdl.K_c:
            self.clear()
            self.update()

    def on_mouse_motion(self, event):
        if event.buttons[0] != 1: return
        w, h = self.size
        dx, dy = map(float, event.rel)
        rx, ry = pi*dx/w, pi*dy/h
        self.drx += rx
        self.dry -= ry
        if time() - self.dt > RATE:
            self.camera.rotate(-self.dry, self.drx)
            self.fast_display()
            self.dt = time()
            self.dry, self.drx = 0.0, 0.0

    def on_mouse_up(self, event):
        if event.button == 1:
            self.dt = 0
            self.dry, self.drx = 0.0, 0.0

    def on_mouse_down(self, event):
        if event.button not in [2,3]: return
        x, y = event.pos

        ray = make_ray(self.camera, (x, y))
        print 'ray:', ray.v
        if not ray.intersect_box(self.obj.box): return

        if event.button is 3:
            t0 = time()
            hit = kd_search(ray, self.obj)
            if hit is not None:
                paint_poly(hit.vectors, self.camera, self.surface, (255, 255, 0))
            else:
                print 'miss'
            print 'smart ray:', time()-t0


    def on_resize(self, event):
        self.camera.resize(event.size)
        self.fast_display()

    def update(self):
        self.painter.blit(self.surface)
        sdl.display.flip()

    def clear(self):
        self.painter.clear()
        self.surface.fill(self.bgcolor)

    def fast_display(self):
        self.clear()
        wireframe_render(self.bbox, self.camera, self.surface, self.fgcolor)
        sdl.display.flip()

    def message(self, text):
        font = sdl.font.Font(None, 16)
        surf = font.render(text, 1, (0, 0, 0), (255, 255, 255))
        dx = surf.get_width()/2
        dy = surf.get_height()
        rect = sdl.Rect(0, 0, self.surface.get_width()-1, dy)
        pos = surf.get_rect(centerx=dx)
        self.surface.fill((255, 255, 255), rect)
        self.surface.blit(surf, pos)
        sdl.display.flip()

    def wireframe(self):
        blue = sdl.color.Color('blue')
        hidden_wireframe_render(self.obj, self.camera, self.surface, blue)
        sdl.display.flip()

    def points(self):
        blue = (0, 0, 255)
        self.painter.clear()
        point_render(self.obj, self.camera, self.painter, blue)
        self.update()

    def depth_map(self):
        wait = lambda p: self.message('%d%% depth map'%(100*p))
        t0 = time()
        self.painter.clear()
        self.dmap = compute_depth_map(self.obj, self.camera, wait)
        draw_gray_map(self.painter, self.dmap)
        self.update()
        self.message('%.2f sec'%(time()-t0))

    def normal_map(self):
        wait = lambda p: self.message('%d%% normal map'%(100*p))
        t0 = time()
        self.painter.clear()
        self.nmap = compute_normal_map(self.obj, self.camera, wait)
        draw_color_map(self.painter, self.nmap)
        self.update()
        self.message('%.2f sec'%(time()-t0))

    def detect_dmap_edges(self):
        self.painter.clear()
        edges = detect_dmap_edges(self.dmap)
        draw_gray_map(self.painter, edges)
        self.update()

    def detect_nmap_edges(self):
        self.painter = self.painter
        edges = detect_nmap_edges(self.nmap, False)
        draw_gray_map(self.painter, edges)
        self.update()

    def combine_dmap_nmap(self):
        self.painter.clear()
        de = detect_dmap_edges(self.dmap)
        ne = detect_nmap_edges(self.nmap, False)
        edges = where(de < ne, de, ne)
        draw_gray_map(self.painter, edges)
        self.update()

    def construct_kd_tree(self):
        t0 = time()
        self.obj.kd_tree = construct_kd_tree(self.obj)
        print 'bsp:', time()-t0

    def compute_silhouette(self):
        self.clear()
        curves = find_silhouette_curves(self.obj, self.camera)
        show_curves(self.obj, curves, self.surface, self.camera.center)

    def screenshot(self):
        filename = 'img%04d.tga'%self.shots
        sdl.image.save(self.surface, filename)
        print 'saved:', filename
        self.shots += 1


try:
    print 'Loading scene...'
    scene = scene_from_mesh(argv[1], (400, 300))
    print len(scene.objects[0].faces), 'triangles'
except IndexError:
    print 'Failed'
    print 'syntax: silhouette.py <file>'
else:
    viewer = SceneViewer(scene, 'Silhouette')
    viewer.start()
